#include "RtAudio.h"
#include "RtMidi.h"
#include "Delay.h"
#include "KSString.h"
#include "KSInstrument.h"
#include <cstdlib>
#include <iostream>
#include <math.h>
#include <signal.h>
using namespace std;

#define SAMPLE double
#define MY_PI 3.14159265358979
#define MY_SRATE 44100.0

//Rtmidi variable and interrupt fn
bool done;
static void finish( int ignore ){ done = true; }


// Callback function
int tick( void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames,
           double streamTime, RtAudioStreamStatus status, void *ksInstrumentPointer )
{
  //output warning if stream status is not good
  if ( status ) cerr << "Stream over/underflow detected." << endl;
  
  //recast output and input buffer pointers
  register SAMPLE *oSamples = (SAMPLE *) outputBuffer;
  register SAMPLE *iSamples = (SAMPLE *) inputBuffer;
	
	//recast string pointer
	KSInstrument* ksInstrument = (KSInstrument*) ksInstrumentPointer;

	for (unsigned int i = 0; i < nBufferFrames; i++)
	{		
		*oSamples++ = ksInstrument->tick();
	}
    
  return 0;
}


int main(int argc, char *argv[])
{
	// Karplus-Strong Instrument variables ------------------------------------------
	int numberOfVoices = 30;		//Maximum number of voices generated at once
	int numberOfStrings = 88;		//Number of strings to populate. All strings to be made accessible must be initialized up front (strings do not change frequency to the demanded notes)
	double lowFreq = 27.5;			//The lowest frequency to be made available on our instrument. All other notes will be generated based on the equal temperment scale, using this as the lowest frequency and generating the specified number of notes.
	double feedbackGain = 0.99;	//Set the feedback gain to be used for all strings
	
	KSInstrument* ksInstrument = new KSInstrument(numberOfStrings, numberOfVoices, lowFreq, feedbackGain, MY_SRATE);
	
	
	// RtMidi Variables -------------------------------------------------------------
	RtMidiIn *midiin = 0;
  std::vector<unsigned char> message;
  int nBytes, i;
  double stamp;
  unsigned int port = 0;
  
  // Rtmidi constructor
  try {
    midiin = new RtMidiIn();
  }
  catch ( RtError &error ) {
    error.printMessage();
    exit( EXIT_FAILURE );
  }
  
  // Don't ignore sysex, timing, or active sensing messages.
  midiin->ignoreTypes( false, false, false );

  // Install an interrupt handler function.
  done = false;
  (void) signal(SIGINT, finish);
	
	
  // RtAudio Variables ------------------------------------------------------------
  unsigned int inChannels = 0, outChannels = 1, fs = MY_SRATE, bufferFrames = 256, iDevice = 0, oDevice = 0, iOffset = 0, oOffset = 0;

  RtAudio adac;
  if ( adac.getDeviceCount() < 1 ) {
    std::cout << "\nNo audio devices found!\n";
    exit( 1 );
  }

  // Let RtAudio print messages to stderr.
  adac.showWarnings( true );

  // Set the number of channels for output.
  RtAudio::StreamParameters oParams;
  
  // Output
  if ( oDevice == 0 )
    oParams.deviceId = adac.getDefaultOutputDevice();
  else
    oParams.deviceId = oDevice - 1;
    
  oParams.nChannels = outChannels;
  oParams.firstChannel = oOffset;

	//Create options struct in case we want to set some of these
  RtAudio::StreamOptions options;

  try {
  		adac.openStream( &oParams, NULL, RTAUDIO_FLOAT64, fs, &bufferFrames, &tick, (void*) ksInstrument, &options );
  }
  catch ( RtError& err ) {
    err.printMessage();
    exit( 1 );
  }

	// Try to open audio stream
  try {
    adac.startStream();
  }
  catch ( RtError& err ) {
    err.printMessage();
    goto cleanup;
  }
  
  // Try to open midi input port
	try {
    midiin->openPort( port );
  }
  catch ( RtError &error ) {
    error.printMessage();
    goto cleanup;
  }


  while ( !done )
  {
    stamp = midiin->getMessage( &message );
    nBytes = message.size();
	
		if (nBytes > 0 )
		{
			//If the message is coming from a key and it is in the range of instantiated notes
			if ( (int)message[0] == 144 && (int)message[1] < numberOfStrings )
			{		
				if ((int)message[2] == 0)	//If we receive a 0 velocity message, stop playing that note
					ksInstrument->markInactive((int)message[1]);
				else	//Otherwise, pluck that string	
					ksInstrument->pluckString((int)message[1], (int)message[2]);	
			}
			else if ( (int)message[0] == 128 )	//Note off message
				ksInstrument->markInactive((int)message[1]);
		}
		
    usleep( 10000 );
  }

  // If we get here, stop the stream.
  try
  {
    adac.stopStream();
  }
  catch( RtError& err )
  {
    err.printMessage();
  }


cleanup:
  if ( adac.isStreamOpen() ) adac.closeStream();
	delete ksInstrument;
	delete midiin;

  return 0;
}
